# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

from pwem.viewers import ObjectView

import pyworkflow.viewer as pwviewer

import tomo.objects
from tomo.viewers.views_tkinter_tree import MeshesTreeProvider

from dynamo.viewers.views_tkinter_tree import DynamoDialog



class DynamoDataViewer(pwviewer.Viewer):
    """ Wrapper to visualize different type of objects
    with the Xmipp program xmipp_showj
    """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [
        tomo.objects.SetOfMeshes
    ]

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)
        self._views = []

    def _getObjView(self, obj, fn, viewParams={}):
        return ObjectView(
            self._project, obj.strId(), fn, viewParams=viewParams)

    def _visualize(self, obj, **kwargs):
        views = []
        cls = type(obj)

        if issubclass(cls, tomo.objects.SetOfMeshes):
            outputMeshes = obj

            meshList = [item.clone() for item in outputMeshes.iterItems()]
            meshList.reverse()

            path = self.protocol._getTmpPath()

            meshProvider = MeshesTreeProvider(meshList,)

            setView = DynamoDialog(self._tkRoot, path, True, provider=meshProvider,)

        return views



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     you (you@yourinstitution.email)
# *
# * your institution
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

"""
@article{CASTANODIEZ2012139,
title = "Dynamo: A flexible, user-friendly development tool for subtomogram averaging of cryo-EM data in high-performance computing environments",
journal = "Journal of Structural Biology",
volume = "178",
number = "2",
pages = "139 - 151",
year = "2012",
note = "Special Issue: Electron Tomography",
issn = "1047-8477",
doi = "https://doi.org/10.1016/j.jsb.2011.12.017",
url = "http://www.sciencedirect.com/science/article/pii/S1047847711003650",
author = "Daniel Castaño-Díez and Mikhail Kudryashev and Marcel Arheit and Henning Stahlberg",
keywords = "Subtomogram averaging, Single Particle Tomography, High-performance computing, GPU computing, Classification",
abstract = "Dynamo is a new software package for subtomogram averaging of cryo Electron Tomography (cryo-ET) data with three main goals: first, Dynamo allows user-transparent adaptation to a variety of high-performance computing platforms such as GPUs or CPU clusters. Second, Dynamo implements user-friendliness through GUI interfaces and scripting resources. Third, Dynamo offers user-flexibility through a plugin API. Besides the alignment and averaging procedures, Dynamo includes native tools for visualization and analysis of results and data, as well as support for third party visualization software, such as Chimera UCSF or EMAN2. As a demonstration of these functionalities, we studied bacterial flagellar motors and showed automatically detected classes with absent and present C-rings. Subtomogram averaging is a common task in current cryo-ET pipelines, which requires extensive computational resources and follows a well-established workflow. However, due to the data diversity, many existing packages offer slight variations of the same algorithm to improve results. One of the main purposes behind Dynamo is to provide explicit tools to allow the user the insertion of custom designed procedures – or plugins – to replace or complement the native algorithms in the different steps of the processing pipeline for subtomogram averaging without the burden of handling parallelization. Custom scripts that implement new approaches devised by the user are integrated into the Dynamo data management system, so that they can be controlled by the GUI or the scripting capacities. Dynamo executables do not require licenses for third party commercial software. Sources, executables and documentation are freely distributed on http://www.dynamo-em.org."
}

"""



# **************************************************************************
# *
# * Authors:     you (you@yourinstitution.email)
# *
# * your institution
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import join
import pwem
import pyworkflow.utils as pwutils
from .constants import *

_logo = "icon.png"
_references = ['CASTANODIEZ2012139']

class Plugin(pwem.Plugin):
    _homeVar = DYNAMO_HOME
    _pathVars = [DYNAMO_HOME]
    # _supportedVersions =

    @classmethod
    def _defineVariables(cls):
        cls._defineEmVar(DYNAMO_HOME, 'dynamo-{}'.format(DYNAMO_VERSION))

    @classmethod
    def getEnviron(cls):
        """ Create the needed environment for Dynamo programs. """
        environ = pwutils.Environ(os.environ)
        dyn_home = cls.getHome()
        environ.update({
            'PATH': dyn_home + '/matlab/bin:' + dyn_home + '/matlab/src:' + dyn_home + '/cuda/bin:' + dyn_home + '/mpi',
            'LD_LIBRARY_PATH': dyn_home + '/MCRLinux/runtime/glnxa64:' + dyn_home + '/MCRLinux/bin/glnxa64:' +
                               dyn_home + '/MCRLinux/sys/os/glnxa64',
            'DYNAMO_ROOT': dyn_home
        }, position=pwutils.Environ.BEGIN)
        return environ

    @classmethod
    def getDynamoProgram(cls):
        return join(cls.getHome(), 'matlab', 'bin', DYNAMO_PROGRAM)

    @classmethod
    def runDynamo(cls, protocol, args, cwd=None):
        """ Run Dynamo command from a given protocol. """
        # args will be the .doc file which contains the MATLAB code
        program = cls.getDynamoProgram()
        protocol.runJob(program, args, env=cls.getEnviron(), cwd=cwd)

    @classmethod
    def defineBinaries(cls, env):
        env.addPackage('dynamo',
                       version='1.146',
                       tar='dynamo-1.146.tar.gz',
                       commands=[],
                       default=True)




# **************************************************************************
# *
# * Authors:  Estrella Fernandez Gimenez (me.fernandez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import math
import numpy as np
from pwem import Domain
from pwem.emlib.image.image_handler import ImageHandler
from pwem.objects.data import Transform
Coordinate3D = Domain.importFromPlugin("tomo.objects", "Coordinate3D")
TomoAcquisition = Domain.importFromPlugin("tomo.objects", "TomoAcquisition")


def writeVolume(volume, outputFn):
    ih = ImageHandler()
    ih.convert(volume, "%s.mrc" % outputFn)


def writeSetOfVolumes(setOfVolumes, outputFnRoot, name):
    ih = ImageHandler()
    if name == 'id':  # write the ID of the object in the name
        for volume in setOfVolumes:
            ih.convert(volume, "%s%03d.mrc" % (outputFnRoot, volume.getObjId()))
    if name == 'ix':  # write the INDEX of the object in the name
        for ix, volume in enumerate(setOfVolumes):
            ih.convert(volume, "%s%03d.mrc" % (outputFnRoot, int(ix+1)))


def writeDynTable(fhTable, setOfSubtomograms):
    for subtomo in setOfSubtomograms.iterItems():
        if subtomo.hasCoordinate3D():
            x = subtomo.getCoordinate3D().getX()
            y = subtomo.getCoordinate3D().getY()
            z = subtomo.getCoordinate3D().getZ()
        else:
            x = 0
            y = 0
            z = 0
        if subtomo.hasTransform():
            tdrot, tilt, narot, shiftx, shifty, shiftz = matrix2eulerAngles(subtomo.getTransform().getMatrix())
        else:
            tilt = 0
            narot = 0
            tdrot = 0
            shiftx = 0
            shifty = 0
            shiftz = 0
        if subtomo.hasAcquisition():
            anglemin = subtomo.getAcquisition().getAngleMin()
            anglemax = subtomo.getAcquisition().getAngleMax()
        else:
            anglemin = 0
            anglemax = 0
        fhTable.write('%d 1 1 %d %d %d %d %d %d 0 0 0 1 %d %d 0 0 0 0 0 0 1 0 %d %d %d 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n'
                      % (subtomo.getObjId(), shiftx, shifty, shiftz, tdrot, tilt, narot, anglemin, anglemax, x, y, z))


def readDynTable(self, item):
    nline = next(self.fhTable)
    nline = nline.rstrip()
    id = int(nline.split()[0])
    item.setObjId(id)
    shiftx = nline.split()[3]
    shifty = nline.split()[4]
    shiftz = nline.split()[5]
    tdrot = nline.split()[6]
    tilt = nline.split()[7]
    narot = nline.split()[8]
    A = eulerAngles2matrix(tdrot, tilt, narot, shiftx, shifty, shiftz)
    transform = Transform()
    transform.setMatrix(A)
    item.setTransform(transform)
    angleMin = nline.split()[13]
    angleMax = nline.split()[14]
    acquisition = TomoAcquisition()
    acquisition.setAngleMin(angleMin)
    acquisition.setAngleMax(angleMax)
    item.setAcquisition(acquisition)
    volId = nline.split()[19]
    x = nline.split()[23]
    y = nline.split()[24]
    z = nline.split()[25]
    coordinate3d = Coordinate3D()
    coordinate3d.setVolId(volId)
    coordinate3d.setX(float(x))
    coordinate3d.setY(float(y))
    coordinate3d.setZ(float(z))
    item.setCoordinate3D(coordinate3d)
    classId = nline.split()[21]
    item.setClassId(classId)


def readDynCoord(tableFile, coord3DSet, tomo):
    with open(tableFile) as fhTable:
        for nline in fhTable:
            coordinate3d = Coordinate3D()
            nline = nline.rstrip()
            shiftx = nline.split()[3]
            shifty = nline.split()[4]
            shiftz = nline.split()[5]
            tdrot = nline.split()[6]
            tilt = nline.split()[7]
            narot = nline.split()[8]
            A = eulerAngles2matrix(tdrot, tilt, narot, shiftx, shifty, shiftz)
            x = nline.split()[23]
            y = nline.split()[24]
            z = nline.split()[25]
            coordinate3d.setX(float(x))
            coordinate3d.setY(float(y))
            coordinate3d.setZ(float(z))
            coordinate3d.setMatrix(A)
            coordinate3d.setVolume(tomo)
            coord3DSet.append(coordinate3d)


# matrix2euler dynamo
def matrix2eulerAngles(A):
    tol = 1e-4
    if abs(A[2, 2] - 1) < tol:
        tilt = 0
        narot = math.atan2(A[1, 0], A[0, 0]) * 180 / math.pi
        tdrot = 0
    elif abs(A[2, 2] + 1) < tol:
        tdrot = 0
        tilt = 180
        narot = math.atan2(A[1, 0], A[0, 0]) * 180 / math.pi
    else:
        tdrot = math.atan2(A[2, 0], A[2, 1])
        tilt = math.acos(A[2, 2])
        narot = math.atan2(A[0, 2], -A[1, 2])
    tilt = tilt * 180 / math.pi
    narot = narot * 180 / math.pi
    tdrot = tdrot * 180 / math.pi
    return tdrot, tilt, narot, A[0, 3], A[1, 3], A[2, 3]


# euler2matrix dynamo
def eulerAngles2matrix(tdrot, tilt, narot, shiftx, shifty, shiftz):
    tdrot = float(tdrot)
    tilt = float(tilt)
    narot = float(narot)
    tdrot = tdrot * math.pi / 180
    narot = narot * math.pi / 180
    tilt = tilt * math.pi / 180
    costdrot = math.cos(tdrot)
    cosnarot = math.cos(narot)
    costilt = math.cos(tilt)
    sintdrot = math.sin(tdrot)
    sinnarot = math.sin(narot)
    sintilt = math.sin(tilt)
    A = np.empty([4, 4])
    A[3, 3] = 1
    A[3, 0:3] = 0
    A[0, 3] = float(shiftx)
    A[1, 3] = float(shifty)
    A[2, 3] = float(shiftz)
    A[0, 0] = costdrot*cosnarot - sintdrot*costilt*sinnarot
    A[0, 1] = -cosnarot*sintdrot - costdrot*costilt*sinnarot
    A[0, 2] = sinnarot*sintilt
    A[1, 0] = costdrot*sinnarot + cosnarot*sintdrot*costilt
    A[1, 1] = costdrot*cosnarot*costilt - sintdrot*sinnarot
    A[1, 2] = -cosnarot*sintilt
    A[2, 0] = sintdrot*sintilt
    A[2, 1] = costdrot*sintilt
    A[2, 2] = costilt
    return A



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os, threading
import numpy as np

import pyworkflow.config as conf
from pyworkflow import utils as pwutils
from pyworkflow.utils.process import runJob
from pyworkflow.gui.dialog import ToolbarListDialog

from dynamo import Plugin


class DynamoDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    an Eman subprocess from a list of Tomograms.
    """

    def __init__(self, parent, path, viewer, **kwargs):
        self.path = path
        self.provider = kwargs.get("provider", None)
        if viewer:
            ToolbarListDialog.__init__(self, parent,
                                       "Tomogram List",
                                       allowsEmptySelection=False,
                                       itemDoubleClick=self.doubleClickViewer,
                                       **kwargs)
        else:
            ToolbarListDialog.__init__(self, parent,
                                       "Tomogram List",
                                        allowsEmptySelection=False,
                                        itemDoubleClick=self.doubleClickOnTomogram,
                                        **kwargs)

    def refresh_gui(self):
        if self.proc.isAlive():
            self.after(1000, self.refresh_gui)
        else:
            self.tree.update()
            pwutils.cleanPath(os.path.join(self.path, 'tomos'))
            pwutils.cleanPath(os.path.join(self.path, 'tomos.ctlg'))

    def refresh_gui_viewer(self):
        if self.proc.isAlive():
            self.after(1000, self.refresh_gui_viewer)
        else:
            pwutils.cleanPath(os.path.join(self.path, 'mesh.txt'))


    def doubleClickOnTomogram(self, e=None):
        self.tomo = e
        self.proc = threading.Thread(target=self.lanchDynamoForTomogram, args=(self.tomo,))
        self.proc.start()
        self.after(1000, self.refresh_gui)

    def doubleClickViewer(self, e=None):
        self.tomo = e
        self.proc = threading.Thread(target=self.lanchDynamoForViewing, args=(self.tomo,))
        self.proc.start()
        self.after(1000, self.refresh_gui_viewer)

    def lanchDynamoForTomogram(self, tomo):
        inputFilePath = self._writeInputFile(tomo)

        args = ' %s' % inputFilePath
        runJob(None, Plugin.getDynamoProgram(), args, env=Plugin.getEnviron())

    def lanchDynamoForViewing(self, mesh):
        self.mesh = mesh
        inputFilePath = self._writeViewerFile(mesh.getVolume())

        args = ' %s' % inputFilePath
        runJob(None, Plugin.getDynamoProgram(), args, env=Plugin.getEnviron())

    def _writeInputFile(self, tomo):
        inputFilePath = os.path.join(conf.Config.SCIPION_TMP, "commands.doc")
        tomoFile = os.path.join(conf.Config.SCIPION_TMP, "tomos.vll")
        catalogue = os.path.join(self.path, 'tomos')
        tomoFid = open(tomoFile, 'w')
        tomoName = tomo.getFileName()
        tomoName = os.path.basename(tomoName)
        tomoFid.write(tomoName + '\n')
        tomoFid.close()
        tomoBase = pwutils.removeBaseExt(tomo.getFileName())
        inputFid = open(inputFilePath, 'w')
        if not os.path.isfile(os.path.join(self.path, tomoBase + '.txt')):
            content = 'dcm -create %s -fromvll %s \n' \
                      'newClass = \'Yes\' \n' \
                      'class = 0 \n' \
                      'while strcmp(newClass, \'Yes\') \n' \
                      'class = class + 1 \n' \
                      'm = dmodels.membraneByLevels()\n' \
                      'm.linkCatalogue(\'%s\',\'i\',1,\'s\',class)\n' \
                      'm.saveInCatalogue()\n' \
                      'dtmslice %s -c %s \n' \
                      'modeltrack.addOne(\'model\',m)\n' \
                      'modeltrack.setActiveModel(class)\n' \
                      'uiwait(dpkslicer.getHandles().figure_fastslicer)\n' \
                      'modeltrack.saveAllInCatalogue\n' \
                      'm = dread(dcmodels(\'%s\',\'i\', 1))\n' \
                      'if class > 1 \n' \
                      'aux = [m{class}.points class*ones(length(m{class}.points), 1)] \n' \
                      'mat = [mat ; aux] \n' \
                      'else \n' \
                      'mat = [m.points class*ones(length(m.points), 1)] \n' \
                      'end \n' \
                      'newClass = questdlg(\'Create a new class\', \'Dialog\', \'Yes\', \'No\', \'Yes\') \n' \
                      'end \n' \
                      'writematrix(mat, \'%s\')\n' \
                      'exit' % (catalogue, tomoFile, catalogue, tomo.getFileName(), catalogue,
                                catalogue, os.path.join(self.path, tomoBase + '.txt'))
            # 'writematrix(m.mesh.tr.ConnectivityList, \'%s\')\n' \
        else:
            content = 'dcm -create %s -fromvll %s\n' \
                      'idx=1\n' \
                      'modelData=readmatrix(\'%s\')\n' \
                      'class=unique(modelData(:,4))\n' \
                      'for item=class\'\n' \
                      'm=dmodels.membraneByLevels()\n' \
                      'logic=repmat(modelData(:,4)==item,1,4)\n' \
                      'data=modelData(logic)\n' \
                      'data=reshape(data,length(data)/4,4)\n' \
                      'addPoint(m,data(:,1:3),data(:,3))\n' \
                      'm.linkCatalogue(\'%s\',\'i\',1,\'s\',item)\n' \
                      'm.saveInCatalogue()\n' \
                      'modeltrack.addOne(\'model\',m)\n' \
                      'end\n' \
                      'newClass=\'Yes\'\n' \
                      'while strcmp(newClass,\'Yes\')\n' \
                      'class=[class;class(end)+1]\n' \
                      'm=dmodels.membraneByLevels()\n' \
                      'm.linkCatalogue(\'%s\',\'i\',1,\'s\',class(end))\n' \
                      'm.saveInCatalogue()\n' \
                      'dtmslice %s -c %s\n' \
                      'modeltrack.addOne(\'model\',m)\n' \
                      'modeltrack.setActiveModel(class(end))\n' \
                      'uiwait(dpkslicer.getHandles().figure_fastslicer)\n' \
                      'modeltrack.saveAllInCatalogue\n' \
                      'm=dread(dcmodels(\'%s\',\'i\', 1))\n' \
                      'if idx==1 \n' \
                      'aux=[m.points class(end)*ones(length(m.points),1)]\n' \
                      'modelData=[modelData;aux] \n' \
                      'idx=idx+1\n' \
                      'else \n' \
                      'aux=[m{idx}.points class(end)*ones(length(m{idx}.points),1)]\n' \
                      'modelData=[modelData;aux] \n' \
                      'idx=idx+1\n' \
                      'end \n' \
                      'newClass=questdlg(\'Create a new class\',\'Dialog\',\'Yes\',\'No\',\'Yes\')\n' \
                      'end\n' \
                      'writematrix(modelData,\'%s\')\n' \
                      'exit' % (catalogue, tomoFile, os.path.join(self.path, tomoBase + '.txt'),
                                catalogue, catalogue, tomo.getFileName(), catalogue,
                                catalogue, os.path.join(self.path, tomoBase + '.txt'))
        inputFid.write(content)
        inputFid.close()
        return inputFilePath

    def _writeViewerFile(self, tomo):
        inputFilePath = os.path.join(conf.Config.SCIPION_TMP, "commands.doc")
        tomoFile = os.path.join(conf.Config.SCIPION_TMP, "tomos.vll")
        catalogue = os.path.join(self.path, 'tomos')
        tomoFid = open(tomoFile, 'w')
        tomoName = tomo.getFileName()
        tomoName = os.path.basename(tomoName)
        tomoFid.write(tomoName + '\n')
        tomoFid.close()
        mesh_group = self.mesh.getMesh()
        group = np.ones((1, len(mesh_group))) * self.mesh.getGroup()
        meshFile = 'mesh.txt'
        np.savetxt(os.path.join(self.path, meshFile), np.append(mesh_group, group.T, axis=1), fmt='%d', delimiter=',')
        inputFid = open(inputFilePath, 'w')
        content = 'dcm -create %s -fromvll %s\n' \
                  'modelData=readmatrix(\'%s\')\n' \
                  'class=unique(modelData(:,4))\n' \
                  'for item=class\'\n' \
                  'm=dmodels.membraneByLevels()\n' \
                  'logic=repmat(modelData(:,4)==item,1,4)\n' \
                  'data=modelData(logic)\n' \
                  'data=reshape(data,length(data)/4,4)\n' \
                  'addPoint(m,data(:,1:3),data(:,3))\n' \
                  'm.linkCatalogue(\'%s\',\'i\',1,\'s\',item)\n' \
                  'm.saveInCatalogue()\n' \
                  'modeltrack.addOne(\'model\',m)\n' \
                  'end\n' \
                  'dtmslice %s -c %s \n' \
                  'uiwait(dpkslicer.getHandles().figure_fastslicer)\n' \
                  'exit' % (catalogue, tomoFile, os.path.join(self.path, meshFile),
                            catalogue, tomo.getFileName(), catalogue)
        inputFid.write(content)
        inputFid.close()
        return inputFilePath

class DynamoTomoDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    an Eman subprocess from a list of Tomograms.
    """

    def __init__(self, parent, path, **kwargs):
        self.path = path
        self.provider = kwargs.get("provider", None)
        ToolbarListDialog.__init__(self, parent,
                                   "Tomogram List",
                                    allowsEmptySelection=False,
                                    itemDoubleClick=self.doubleClickOnTomogram,
                                    **kwargs)

    def refresh_gui(self):
        if self.proc.isAlive():
            self.after(1000, self.refresh_gui)
        else:
            self.tree.update()

    def doubleClickOnTomogram(self, e=None):
        self.tomo = e
        self.proc = threading.Thread(target=self.lanchDynamoForTomogram, args=(self.tomo,))
        self.proc.start()
        self.after(1000, self.refresh_gui)

    def lanchDynamoForTomogram(self, tomo):
        commandsFile = self.writeMatlabCode(tomo)
        args = ' %s' % commandsFile
        runJob(None, Plugin.getDynamoProgram(), args, env=Plugin.getEnviron())

    def writeMatlabCode(self, tomo):
        # Initialization params
        codeFilePath = os.path.join(os.getcwd(), "DynamoPicker.m")
        catalogue = os.path.join(self.path, "tomos")

        # Write code to Matlab code file
        codeFid = open(codeFilePath, 'w')
        content = "catalogue_name='%s'\n" \
                  "c=dread(strcat(catalogue_name,'.ctlg'))\n" \
                  "n=cellfun(@(c) c.fullFileName,c.volumes,'UniformOutput',false)\n" \
                  "idt=find(cell2mat(cellfun(@(c) strcmp(c,'%s'),n,'UniformOutput',false)))\n" \
                  "dtmslice %s -c %s \n" \
                  "modeltrack.loadFromCatalogue('handles',c,'full',true,'select',false)\n" \
                  "uiwait(dpkslicer.getHandles().figure_fastslicer)\n" \
                  "models = dread(dcmodels(catalogue_name,'i', idt))\n" \
                  "outFile='%s'\n" \
                  "savePath='%s'\n" \
                  "outPoints=[outFile '.txt']\n" \
                  "outAngles=['angles_' outFile '.txt']\n" \
                  "crop_points=[]\n" \
                  "crop_angles=[]\n" \
                  "for model=models\n" \
                  "if iscell(model)\n" \
                  "crop_points=[crop_points; model{end}.crop_points]\n" \
                  "crop_angles=[crop_angles; model{end}.crop_angles]\n" \
                  "else\n" \
                  "crop_points=[crop_points; model.crop_points]\n" \
                  "crop_angles=[crop_angles; model.crop_angles]\n" \
                  "end\n" \
                  "end\n" \
                  "if ~isempty(crop_points)\n" \
                  "writematrix(crop_points,fullfile(savePath,outPoints),'Delimiter',' ')\n" \
                  "writematrix(crop_angles,fullfile(savePath,outAngles),'Delimiter',' ')\n" \
                  "end\n" \
                  'exit\n' % (catalogue, tomo.getFileName(), tomo.getFileName(), catalogue,
                            pwutils.removeBaseExt(tomo.getFileName()), self.path)
        codeFid.write(content)
        codeFid.close()

        return codeFilePath


# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os, glob
import numpy as np

from pwem.protocols import ProtAnalysis3D

from pyworkflow.protocol.params import PointerParam
import pyworkflow.utils as pwutils
from pyworkflow.utils.properties import Message

from dynamo.viewers.views_tkinter_tree import DynamoDialog

from tomo.viewers.views_tkinter_tree import TomogramsTreeProvider
from tomo.objects import Mesh
from tomo.protocols import ProtTomoBase


"""
Protocols to create models in Dynamo
"""

class DynamoModels(ProtAnalysis3D, ProtTomoBase):
    """Model manger from Dynamo for Mesh creation.
    After opening the desired Tomogram, a surface model will be created and loaded in Dynamo automatically.
    Once the points of the Mesh have been defined, close Dynamo to save automatically your data.
    If a dialog asking to keep the models loaded in memory appears, click on 'Keep it!!' or close the dialog."""
    _label = 'model manager'

    def __init__(self, **args):
        ProtAnalysis3D.__init__(self, **args)
        self.OUTPUT_PREFIX = 'outputMeshes'

        # --------------------------- DEFINE param functions ------------------------
    def _defineParams(self, form):
        form.addSection(label='Input Parameters')
        form.addParam('inputTomograms', PointerParam,
                      pointerClass="SetOfTomograms",
                      label='Set of tomograms',
                      help="Set of tomograms to create a model")

        # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('launchDynamoGUIStep', interactive=True)

        # --------------------------- STEPS functions -------------------------------
    def launchDynamoGUIStep(self):
        tomoList = [tomo.clone() for tomo in self.inputTomograms.get().iterItems()]

        tomoProvider = TomogramsTreeProvider(tomoList, self._getExtraPath(), "txt")

        self.dlg = DynamoDialog(None, self._getExtraPath(), False, provider=tomoProvider,)

        if glob.glob(self._getExtraPath("*.txt")):
            self._createOutput()

    def _createOutput(self):
        outSet = self._createSetOfMeshes()
        for tomo in self.inputTomograms.get().iterItems():
            tomoName = pwutils.removeBaseExt(tomo.getFileName())
            outFile = self._getExtraPath(tomoName + '.txt')
            if os.path.isfile(outFile):
                data = np.loadtxt(outFile, delimiter=',')
                groups = np.unique(data[:, 3]).astype(int)
                for group in groups:
                    mesh = Mesh(group=group, path=outFile)
                    mesh.setVolume(tomo.clone())
                    outSet.append(mesh)
        outSet.setVolumes(self.inputTomograms.get())
        name = self.OUTPUT_PREFIX
        args = {name: outSet}
        self._defineOutputs(**args)
        self._defineSourceRelation(self.inputTomograms.get(), outSet)

        # Update Outputs
        outSet.setObjComment(self.getSummary())
        self._updateOutputSet(name, outSet, state=outSet.STREAM_CLOSED)

        # --------------------------- INFO functions --------------------------------
    def _summary(self):
        if self.getOutputsSize() >= 1:
            summary = []
            for key, output in self.iterOutputAttributes():
                summary.append("*%s:* \n %s \n" % (key, output.getObjComment()))
        else:
            summary = []
            summary.append(Message.TEXT_NO_OUTPUT_CO)

        return summary

    def getSummary(self):
        summary = []
        count = 0
        for file in os.listdir(self._getExtraPath()):
            if file.endswith(".txt"):
                count += 1
        summary.append("Meshes defined for %d/%d files have been saved in Scipion (%s)" % (
            count/2, self.inputTomograms.get().getSize(), self._getExtraPath()))

        return "\n".join(summary)

    def _methods(self):
        tomos = self.inputTomograms.get()
        return [
            "Model creation using Dynamo",
            "A total of %d tomograms of dimensions %s were used"
            % (tomos.getSize(), tomos.getDimensions()),
        ]

    def _citations(self):
        return ['CASTANODIEZ2012139']



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Estrella Fernandez Gimenez (me.fernandez@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.tests import BaseTest, setupTestProject
from tomo.protocols import ProtImportSubTomograms, ProtImportCoordinates3D, ProtImportTomograms
from tomo.tests import DataSet
from xmipp3.protocols import XmippProtCreateMask3D
from dynamo.protocols import DynamoSubTomoMRA, DynamoExtraction


class TestDynamoBase(BaseTest):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)

    @classmethod
    def setData(cls, projectData='tomo-em'):
        cls.dataset = DataSet.getDataSet(projectData)
        cls.tomogram = cls.dataset.getFile('tomo1')
        cls.emanCoords = cls.dataset.getFile('overview_wbp.txt')
        cls.dynCoords = cls.dataset.getFile('overview_wbp.tbl')
        cls.angles = cls.dataset.getFile('overview_wbp.ang')
        cls.inputSetOfSubTomogram = cls.dataset.getFile('subtomo')
        cls.smallTomogram = cls.dataset.getFile('coremask_normcorona.mrc')

class TestSubTomogramsAlignment(BaseTest):
    """ This class check if the protocol to import sub tomograms works
    properly."""

    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        cls.dataset = DataSet.getDataSet('tomo-em')
        cls.setOfSubtomograms = cls.dataset.getFile('basename.hdf')

    def _runPreviousProtocols(self):
        protImport = self.newProtocol(ProtImportSubTomograms,
                                      filesPath=self.setOfSubtomograms,
                                      samplingRate=5)
        self.launchProtocol(protImport)
        return protImport

    def _runAlignment(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     generateTemplate=True)
        alignment.templateRef.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.averageSubTomogram,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithTemplate(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     templateRef=protImport)
        alignment.templateRef.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.averageSubTomogram,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithTemplatesMRA(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     templateRef=particles,
                                     mra=True,
                                     fmask=particles)
        alignment.fmask.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.AverageRef1,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithoutTemplatesMRA(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     mra=True,
                                     generateTemplate=True,
                                     nref=4)
        alignment.fmask.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.AverageRef1,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithMask(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        protMask = self.newProtocol(XmippProtCreateMask3D,
                                    inputVolume=protImport,
                                    source=0,
                                    volumeOperation=0,
                                    threshold=0.4)
        protMask.inputVolume.setExtended("outputSubTomograms.1")
        protMask.setObjLabel('threshold mask')
        self.launchProtocol(protMask)
        self.assertIsNotNone(protMask.outputMask,
                             "There was a problem with create mask from volume")
        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     templateRef=protImport,
                                     mask=protMask.outputMask)
        alignment.templateRef.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.averageSubTomogram,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithMasks(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        protMask = self.newProtocol(XmippProtCreateMask3D,
                                    inputVolume=protImport,
                                    source=0, volumeOperation=0,
                                    threshold=0.4)
        protMask.inputVolume.setExtended("outputSubTomograms.1")
        protMask.setObjLabel('threshold mask')
        self.launchProtocol(protMask)
        self.assertIsNotNone(protMask.outputMask,
                             "There was a problem with create mask from volume")

        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     templateRef=protImport,
                                     mask=protMask.outputMask,
                                     cmask=protMask.outputMask,
                                     fmask=protMask.outputMask,
                                     smask=protMask.outputMask)
        alignment.templateRef.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.averageSubTomogram,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def _runAlignmentWithFmask(self):
        protImport = self._runPreviousProtocols()
        particles = protImport.outputSubTomograms
        protMask = self.newProtocol(XmippProtCreateMask3D,
                                    inputVolume=protImport,
                                    source=0, volumeOperation=0,
                                    threshold=0.4)
        protMask.inputVolume.setExtended("outputSubTomograms.1")
        protMask.setObjLabel('threshold mask')
        self.launchProtocol(protMask)
        self.assertIsNotNone(protMask.outputMask,
                             "There was a problem with create mask from volume")

        alignment = self.newProtocol(DynamoSubTomoMRA,
                                     inputVolumes=particles,
                                     generateTemplate=True,
                                     fmask=protMask.outputMask)
        alignment.templateRef.setExtended("outputSubTomograms.1")
        self.launchProtocol(alignment)
        self.assertIsNotNone(alignment.averageSubTomogram,
                             "There was a problem with SetOfSubtomograms output")
        return alignment

    def test_basicAlignment(self):
        dynamoAlignment = self._runAlignment()
        averageSubTomogram = getattr(dynamoAlignment, 'averageSubTomogram')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomograms')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithTemplate(self):
        dynamoAlignment = self._runAlignmentWithTemplate()
        averageSubTomogram = getattr(dynamoAlignment, 'averageSubTomogram')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomograms')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithMask(self):
        dynamoAlignment = self._runAlignmentWithMask()
        averageSubTomogram = getattr(dynamoAlignment, 'averageSubTomogram')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomograms')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithFask(self):
        dynamoAlignment = self._runAlignmentWithFmask()
        averageSubTomogram = getattr(dynamoAlignment, 'averageSubTomogram')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomograms')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithMasks(self):
        dynamoAlignment = self._runAlignmentWithMasks()
        averageSubTomogram = getattr(dynamoAlignment, 'averageSubTomogram')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomograms')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithTemplatesMRA(self):
        dynamoAlignment = self._runAlignmentWithTemplatesMRA()
        averageSubTomogram = getattr(dynamoAlignment, 'AverageRef1')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomogramsRef1')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

    def test_alignmentWithoutTemplatesMRA(self):
        dynamoAlignment = self._runAlignmentWithoutTemplatesMRA()
        averageSubTomogram = getattr(dynamoAlignment, 'AverageRef1')
        outputSubtomograms = getattr(dynamoAlignment, 'outputSubtomogramsRef1')
        self.assertTrue(averageSubTomogram)
        self.assertTrue(outputSubtomograms)
        self.assertTrue(outputSubtomograms.getFirstItem().hasTransform())
        return dynamoAlignment

class TestDynamoExtraction(TestDynamoBase):
    '''This class checks if the protocol to extract subtomograms
    works properly'''

    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestDynamoBase.setData()

    def _runExtraction(self, tomoSource = 0, downFactor = 1, text=''):

        protImportTomogram = self.newProtocol(ProtImportTomograms,
                                              filesPath=self.tomogram,
                                              samplingRate=5)

        self.launchProtocol(protImportTomogram)

        self.assertIsNotNone(protImportTomogram.outputTomograms,
                             "There was a problem with tomogram output")

        protImportCoordinatesNoAngles = self.newProtocol(ProtImportCoordinates3D,
                                                         objLabel='Coordinates without Angles',
                                                         auto=ProtImportCoordinates3D.IMPORT_FROM_EMAN,
                                                         filesPath=self.emanCoords,
                                                         importTomograms=protImportTomogram.outputTomograms,
                                                         filesPattern='', boxSize=32,
                                                         samplingRate=5)

        protImportCoordinatesAngles = self.newProtocol(ProtImportCoordinates3D,
                                                       objLabel='Coordinates with Angles',
                                                       auto=ProtImportCoordinates3D.IMPORT_FROM_EMAN,
                                                       filesPath=self.dynCoords,
                                                       importTomograms=protImportTomogram.outputTomograms,
                                                       filesPattern='', boxSize=32, importAngles=True,
                                                       samplingRate=5)

        self.launchProtocol(protImportCoordinatesNoAngles)
        self.launchProtocol(protImportCoordinatesAngles)

        coordsNoAngle = protImportCoordinatesNoAngles.outputCoordinates
        coordsAngle = protImportCoordinatesAngles.outputCoordinates

        self.assertIsNotNone(coordsNoAngle,
                             "There was a problem with coordinates 3d output")
        self.assertIsNotNone(coordsAngle,
                             "There was a problem with coordinates 3d output")

        protDynamoExtractionNoAngles = self.newProtocol(DynamoExtraction,
                                                        objLabel='Extraction without Angles - %s' % text,
                                                        inputTomograms=protImportTomogram.outputTomograms,
                                                        inputCoordinates=coordsNoAngle,
                                                        boxSize=coordsNoAngle.getBoxSize(),
                                                        tomoSource=tomoSource,
                                                        downFactor=downFactor)

        protDynamoExtractionAngles = self.newProtocol(DynamoExtraction,
                                                      objLabel='Extraction with Angles - %s' % text,
                                                      inputTomograms=protImportTomogram.outputTomograms,                                                      inputCoordinates=coordsAngle,
                                                      boxSize=coordsAngle.getBoxSize(),
                                                      tomoSource=tomoSource,
                                                      downFactor=downFactor)

        self.launchProtocol(protDynamoExtractionNoAngles)
        self.launchProtocol(protDynamoExtractionAngles)
        return protDynamoExtractionNoAngles, protDynamoExtractionAngles

    def test_Extraction_SameAsPicking(self):
        protExtraction = self._runExtraction(text='SameAsPicking')

        outputNoAngles = getattr( protExtraction[0], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputNoAngles)
        self.assertTrue(outputNoAngles.hasCoordinates3D())
        self.assertTrue(outputNoAngles.getCoordinates3D().getObjValue())

        outputAngles = getattr( protExtraction[1], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputAngles)
        self.assertTrue(outputAngles.hasCoordinates3D())
        self.assertTrue(outputAngles.getCoordinates3D().getObjValue())

        return protExtraction

    def test_Extraction_Other(self):
        protExtraction = self._runExtraction(tomoSource = 1, text='OtherSource')

        outputNoAngles = getattr(protExtraction[0], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputNoAngles)
        self.assertTrue(outputNoAngles.hasCoordinates3D())
        self.assertTrue(outputNoAngles.getCoordinates3D().getObjValue())

        outputAngles = getattr(protExtraction[1], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputAngles)
        self.assertTrue(outputAngles.hasCoordinates3D())
        self.assertTrue(outputAngles.getCoordinates3D().getObjValue())

        return protExtraction

    def test_Extraction_DownSampling(self):
        protExtraction = self._runExtraction(downFactor = 2, text='DownSampling')

        outputNoAngles = getattr(protExtraction[0], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputNoAngles)
        self.assertTrue(outputNoAngles.hasCoordinates3D())
        self.assertTrue(outputNoAngles.getCoordinates3D().getObjValue())

        outputAngles = getattr(protExtraction[1], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputAngles)
        self.assertTrue(outputAngles.hasCoordinates3D())
        self.assertTrue(outputAngles.getCoordinates3D().getObjValue())

        return protExtraction

    def test_Extraction_All(self):
        protExtraction = self._runExtraction(tomoSource = 1, downFactor = 2, text='AllOptions')

        outputNoAngles = getattr(protExtraction[0], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputNoAngles)
        self.assertTrue(outputNoAngles.hasCoordinates3D())
        self.assertTrue(outputNoAngles.getCoordinates3D().getObjValue())

        outputAngles = getattr(protExtraction[1], 'outputSetOfSubtomogram', None)
        self.assertTrue(outputAngles)
        self.assertTrue(outputAngles.hasCoordinates3D())
        self.assertTrue(outputAngles.getCoordinates3D().getObjValue())

        return protExtraction


# **************************************************************************
# *
# * Authors:    David Herreros Calero (dherreros@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import numpy as np

import pyworkflow.utils as pwutils
from pyworkflow.protocol.params import PointerParam, IntParam, EnumParam
from pyworkflow.utils.properties import Message
from pyworkflow.gui.dialog import askYesNo

from tomo.protocols import ProtTomoPicking
from tomo.objects import SetOfCoordinates3D, Coordinate3D
from tomo.viewers.views_tkinter_tree import TomogramsTreeProvider

from dynamo import Plugin
from dynamo.viewers.views_tkinter_tree import DynamoTomoDialog
from dynamo.convert import eulerAngles2matrix

class DynamoBoxing(ProtTomoPicking):
    """Manual vectorial picker from Dynamo. After choosing the Tomogram to be picked, the tomo slicer from Dynamo will be
    direclty loaded with all the models previously saved in the disk (if any).
    After picking, it is needed to go to:
    Active Model > Step-by-step workflow for cropping geometry
    And click -run all- button before closing the window"""

    _label = 'vectorial picking'

    def __init__(self, **kwargs):
        ProtTomoPicking.__init__(self, **kwargs)

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        ProtTomoPicking._defineParams(self, form)

        form.addParam('boxSize', IntParam, label="Box Size")
        form.addParam('selection', EnumParam, choices=['Yes', 'No'], default=1,
                      label='Modify previous coordinates?', display=EnumParam.DISPLAY_HLIST,
                      help='This option allows to add and/or remove coordinates to a previous SetOfCoordinates')
        form.addParam('inputCoordinates', PointerParam, label="Input Coordinates", condition='selection == 0',
                      allowsNull=True, pointerClass='SetOfCoordinates3D',
                      help='Select the previous SetOfCoordinates you want to modify')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        # Copy input coordinates to Extra Path
        self._insertFunctionStep('copyInputCoords')

        # Launch Boxing GUI
        self._insertFunctionStep('launchDynamoBoxingStep', interactive=True)

    # --------------------------- STEPS functions -----------------------------

    def copyInputCoords(self):
        # Initialize the catalogue
        listTomosFile = self._getExtraPath("tomos.vll")
        catalogue = os.path.abspath(self._getExtraPath("tomos"))

        # Create list of tomos file
        tomoFid = open(listTomosFile, 'w')
        for tomo in self.inputTomograms.get().iterItems():
            tomoPath = os.path.abspath(tomo.getFileName())
            tomoFid.write(tomoPath + '\n')
        tomoFid.close()

        # Save coordinates into .txt file for each tomogram
        codeFile = self._getExtraPath('coords2model.m')
        if self.selection.get() == 0:
            inputCoordinates = self.inputCoordinates.get()
            inputTomograms = self.inputTomograms.get()
            for tomo in inputTomograms:
                outFileCoord = self._getExtraPath(pwutils.removeBaseExt(tomo.getFileName())) + ".txt"
                coords_tomo = []
                for coord in inputCoordinates.iterCoordinates(tomo):
                    coords_tomo.append(coord.getPosition())
                if coords_tomo:
                    np.savetxt(outFileCoord, np.asarray(coords_tomo), delimiter=' ')

            # Create small program to tell Dynamo to save the inputCoordinates in a Vesicle Model
            contents = "dcm -create %s -fromvll %s\n" \
                       "path='%s'\n" \
                       "catalogue=dread(['%s' '.ctlg'])\n" \
                       "nVolumes=length(catalogue.volumes)\n" \
                       "for idv=1:nVolumes\n" \
                       "tomoPath=catalogue.volumes{idv}.fullFileName()\n" \
                       "tomoIndex=catalogue.volumes{idv}.index\n" \
                       "[~,tomoName,~]=fileparts(tomoPath)\n" \
                       "coordFile=[path '/' tomoName '.txt']\n" \
                       "if ~isfile(coordFile)\n" \
                       "continue\n" \
                       "end\n" \
                       "coords=readmatrix(coordFile,'Delimiter',' ')\n" \
                       "vesicle=dmodels.vesicle()\n" \
                       "addPoint(vesicle,coords(:,1:3),coords(:,3))\n" \
                       "vesicle.linkCatalogue('%s','i',tomoIndex, 's', 1)\n" \
                       "vesicle.saveInCatalogue()\n" \
                       "end\n" \
                       "exit\n" % (catalogue, listTomosFile, self._getExtraPath(), catalogue, catalogue)
        else:
            contents = "dcm -create %s -fromvll %s\n" % (catalogue, listTomosFile)

        codeFid = open(codeFile, 'w')
        codeFid.write(contents)
        codeFid.close()

        # Tell Dynamo to create the catalogue with the models
        args = ' %s' % codeFile
        self.runJob(Plugin.getDynamoProgram(), args, env=Plugin.getEnviron())

    def launchDynamoBoxingStep(self):

        tomoList = [tomo.clone() for tomo in self.inputTomograms.get().iterItems()]

        tomoProvider = TomogramsTreeProvider(tomoList, self._getExtraPath(), "txt")

        self.dlg = DynamoTomoDialog(None, self._getExtraPath(), provider=tomoProvider)

        # Open dialog to request confirmation to create output
        import tkinter as tk
        if askYesNo(Message.TITLE_SAVE_OUTPUT, Message.LABEL_SAVE_OUTPUT, tk.Frame()):
            self._createOutput()

        pwutils.cleanPattern(self._getExtraPath('*.m'))

    def _createOutput(self):
        coord3DSetDict = {}
        setTomograms = self.inputTomograms.get()
        suffix = self._getOutputSuffix(SetOfCoordinates3D)
        coord3DSet = self._createSetOfCoordinates3D(setTomograms, suffix)
        coord3DSet.setName("tomoCoord")
        coord3DSet.setPrecedents(setTomograms)
        coord3DSet.setSamplingRate(setTomograms.getSamplingRate())
        coord3DSet.setBoxSize(self.boxSize.get())
        for tomo in setTomograms.iterItems():
            outPoints = pwutils.join(self._getExtraPath(), pwutils.removeBaseExt(tomo.getFileName()) + '.txt')
            outAngles = pwutils.join(self._getExtraPath(), 'angles_' + pwutils.removeBaseExt(tomo.getFileName()) + '.txt')
            if not os.path.isfile(outPoints):
                continue

            # Populate Set of 3D Coordinates with 3D Coordinates
            points = np.loadtxt(outPoints, delimiter=' ')
            angles = np.deg2rad(np.loadtxt(outAngles, delimiter=' '))
            for idx in range(len(points)):
                coord = Coordinate3D()
                coord.setPosition(points[idx, 0], points[idx, 1], points[idx, 2])
                matrix = eulerAngles2matrix(angles[idx, 0], angles[idx, 1], angles[idx, 2], 0, 0, 0)
                coord.setMatrix(matrix)
                coord.setVolume(tomo)
                coord3DSet.append(coord)

            coord3DSetDict['00'] = coord3DSet

        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = coord3DSet
        self._defineOutputs(**args)
        self._defineSourceRelation(setTomograms, coord3DSet)
        self._updateOutputSet(name, coord3DSet, state=coord3DSet.STREAM_CLOSED)

    # --------------------------- DEFINE info functions ----------------------
    def getMethods(self, output):
        msg = 'User picked %d particles ' % output.getSize()
        msg += 'with a particle size of %s.' % output.getBoxSize()
        return msg

    def _methods(self):
        methodsMsgs = []
        if self.inputTomograms is None:
            return ['Input tomogram not available yet.']

        methodsMsgs.append("Input tomograms imported of dims %s." %(
                              str(self.inputTomograms.get().getDim())))

        if self.getOutputsSize() >= 1:
            for key, output in self.iterOutputAttributes():
                msg = self.getMethods(output)
                methodsMsgs.append("%s: %s" % (self.getObjectTag(output), msg))
        else:
            methodsMsgs.append(Message.TEXT_NO_OUTPUT_CO)

        return methodsMsgs




# **************************************************************************
# *
# * Authors:    David Herreros Calero (dherreros@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import glob

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params
from pwem.emlib.image import ImageHandler
from pwem.objects import Transform
from pwem.protocols import EMProtocol

from tomo.protocols import ProtTomoBase
from tomo.objects import SetOfSubTomograms, SubTomogram, TomoAcquisition

from dynamo import Plugin
from dynamo.convert import matrix2eulerAngles


# Tomogram type constants for particle extraction
SAME_AS_PICKING = 0
OTHER = 1


class DynamoExtraction(EMProtocol, ProtTomoBase):
    """Extraction of subtomograms using Dynamo"""

    _label = 'vectorial extraction'

    def __init__(self, **kwargs):
        EMProtocol.__init__(self, **kwargs)

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputCoordinates', params.PointerParam, label="Input Coordinates", important=True,
                      pointerClass='SetOfCoordinates3D', help='Select the SetOfCoordinates3D.')
        form.addParam('tomoSource', params.EnumParam,
                      choices=['same as picking', 'other'],
                      default=0,
                      display=params.EnumParam.DISPLAY_HLIST,
                      label='Tomogram source',
                      help='By default the subtomograms will be extracted '
                           'from the tomogram used in the picking '
                           'step ( _same as picking_ option ). \n'
                           'If you select _other_ option, you must provide '
                           'a different tomogram to extract from. \n'
                           '*Note*: In the _other_ case, ensure that provided '
                           'tomogram and coordinates are related ')

        form.addParam('inputTomograms', params.PointerParam,
                      pointerClass='SetOfTomograms',
                      condition='tomoSource != %s' % SAME_AS_PICKING,
                      label='Input tomogram',
                      help='Select the tomogram from which to extract.')

        form.addParam('boxSize', params.FloatParam,
                      label='Box size',
                      help='The subtomograms are extracted as a cubic box of this size.\n'
                           'The wizard selects same box size as picking.\n'
                           'Dynamo only accepts even numbers as box sizes. Otherwise, an exception will '
                           'be returned.')

        form.addParam('downFactor', params.FloatParam, default=1.0,
                      label='Downsampling factor',
                      help='If 1.0 is used, no downsample is applied. '
                           'Non-integer downsample factors are possible. ')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('writeSetOfCoordinates3D')
        self._insertFunctionStep('launchDynamoExtractStep')
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions -----------------------------
    def writeSetOfCoordinates3D(self):
        samplingRateCoord = self.inputCoordinates.get().getSamplingRate()
        samplingRateTomo = self.getInputTomograms().getFirstItem().getSamplingRate()
        self.factor = float(samplingRateCoord / samplingRateTomo)
        self.lines = []
        inputSet = self.getInputTomograms()
        self.coordsFileName = self._getExtraPath('coords.txt')
        self.anglesFileName = self._getExtraPath('angles.txt')
        outC = open(self.coordsFileName, "w")
        outA = open(self.anglesFileName, 'w')
        for idt, item in enumerate(inputSet):
            coordDict = []
            tomo = item.clone()
            for coord3DSet in self.inputCoordinates.get().iterCoordinates():
                if pwutils.removeBaseExt(tomo.getFileName()) == pwutils.removeBaseExt(coord3DSet.getVolName()):
                    angles_coord = matrix2eulerAngles(coord3DSet.getMatrix())
                    x = round(self.factor * coord3DSet.getX())
                    y = round(self.factor * coord3DSet.getY())
                    z = round(self.factor * coord3DSet.getZ())
                    outC.write("%d\t%d\t%d\t%d\n" % (x, y, z, idt+1))
                    outA.write("%f\t%f\t%f\n" % (angles_coord[0], angles_coord[1], angles_coord[2]))
                    coordDict.append(coord3DSet.clone())
            if coordDict:
                self.lines.append(coordDict)
        outC.close()
        outA.close()

    def launchDynamoExtractStep(self):
        codeFilePath = self.writeMatlabCode()

        args = ' %s' % codeFilePath
        self.runJob(Plugin.getDynamoProgram(), args, env=Plugin.getEnviron())

        pwutils.cleanPattern('*.m')

    def createOutputStep(self):
        self.outputSubTomogramsSet = self._createSetOfSubTomograms(self._getOutputSuffix(SetOfSubTomograms))
        self.outputSubTomogramsSet.setSamplingRate(self.getInputTomograms().getSamplingRate() * self.downFactor.get())
        self.outputSubTomogramsSet.setCoordinates3D(self.inputCoordinates)
        acquisition = TomoAcquisition()
        acquisition.setAngleMin(self.getInputTomograms().getFirstItem().getAcquisition().getAngleMin())
        acquisition.setAngleMax(self.getInputTomograms().getFirstItem().getAcquisition().getAngleMax())
        acquisition.setStep(self.getInputTomograms().getFirstItem().getAcquisition().getStep())
        self.outputSubTomogramsSet.setAcquisition(acquisition)
        for ind, tomoFile in enumerate(self.tomoFiles):
            cropPath = os.path.join(self._getExtraPath('Crop%d' % (ind+1)), '')
            if os.path.isdir(cropPath):
                coordSet = self.lines[ind]
                outputSet = self.readSetOfSubTomograms(cropPath, self.outputSubTomogramsSet, coordSet)

        self._defineOutputs(outputSetOfSubtomogram=outputSet)
        self._defineSourceRelation(self.inputCoordinates, outputSet)

    # --------------------------- DEFINE utils functions ----------------------
    def getInputTomograms(self):
        """ Return the tomogram associated to the 'SetOfCoordinates3D' or 'Other' tomograms. """
        if self.tomoSource.get() == SAME_AS_PICKING:
            return self.inputCoordinates.get().getPrecedents()
        else:
            return self.inputTomograms.get()

    def writeMatlabCode(self):
        # Initialization params
        codeFilePath = os.path.join(os.getcwd(), "DynamoExtraction.m")
        listTomosFile = self._getTmpPath("tomos.vll")
        catalogue = os.path.abspath(self._getExtraPath("tomos"))
        self.tomoFiles = [tomo.getFileName() for tomo in self.getInputTomograms().iterItems()]

        # Create list of tomos file
        tomoFid = open(listTomosFile, 'w')
        for tomoFile in self.tomoFiles:
            tomoPath = os.path.abspath(tomoFile)
            tomoFid.write(tomoPath + '\n')
        tomoFid.close()

        # Check of box size is an even number (as required by Dynamo)
        boxSize = round(self.factor * self.boxSize.get())
        if boxSize % 2 != 0:
            boxSize += 1

        # Write code to Matlab code file
        codeFid = open(codeFilePath, 'w')
        content = "path='%s'\n" \
                  "savePath = '%s'\n" \
                  "catalogue_name='%s'\n" \
                  "box=%d\n" \
                  "dcm -create %s -fromvll %s\n" \
                  "c=dread(strcat(catalogue_name,'.ctlg'))\n" \
                  "aux=readmatrix('%s')\n" \
                  "angles=readmatrix('%s')\n" \
                  "coords=aux(:,1:3)\n" \
                  "tags=aux(:,4)'\n" \
                  "for tag=unique(tags)\n" \
                  "tomoCoords=coords(tags==tag,:)\n" \
                  "tomoAngles=angles(tags==tag,:)\n" \
                  "t=dynamo_table_blank(size(tomoCoords,1),'r',tomoCoords,'angles',rad2deg(tomoAngles))\n" \
                  "dtcrop(c.volumes{tag}.fullFileName,t,strcat(savePath,num2str(tag)),box)\n" \
                  "end\n" \
                   % (os.path.abspath(os.getcwd()), self._getExtraPath('Crop'),
                      catalogue,boxSize, catalogue, listTomosFile, os.path.abspath(self.coordsFileName),
                      os.path.abspath(self.anglesFileName))

        codeFid.write(content)
        codeFid.close()

        return codeFilePath

    def readSetOfSubTomograms(self, workDir, outputSubTomogramsSet, coordSet):
        for ids, subTomoFile in enumerate(glob.glob(os.path.join(workDir, '*.em'))):
            subtomogram = SubTomogram()
            subtomogram.cleanObjId()
            subtomogram.setLocation(subTomoFile)
            dfactor = self.downFactor.get()
            if dfactor != 1:
                fnSubtomo = self._getExtraPath(os.path.basename(workDir.strip("/")) + "_downsampled_subtomo%d.mrc" % (ids+1))
                ImageHandler.scaleSplines(subtomogram.getLocation(), fnSubtomo, dfactor)
                subtomogram.setLocation(fnSubtomo)
            subtomogram.setCoordinate3D(coordSet[ids])
            transform = Transform()
            transform.setMatrix(coordSet[ids].getMatrix())
            subtomogram.setTransform(transform)
            outputSubTomogramsSet.append(subtomogram)
        return outputSubTomogramsSet

    # --------------------------- DEFINE info functions ----------------------
    def _validate(self):
        errors = []
        if not self.boxSize.get() % 2 == 0:
            errors.append('The box size provided to Dynamo is an odd number. '
                          'This will rise an execption when Dynamo is invoked.\n'
                          'Please, provide an even number to continue with the extraction.')
        return errors

    def _tomosOther(self):
        """ Return True if other tomograms are used for extract. """
        return self.tomoSource == OTHER

    def _methods(self):
        methodsMsgs = []
        if self.getOutputsSize() >= 1:
            msg = ("A total of %s subtomograms of size %s were extracted"
                   % (str(self.inputCoordinates.get().getSize()), self.boxSize.get()))
            if self._tomosOther():
                msg += (" from another set of tomograms: %s"
                        % self.getObjectTag('inputTomogram'))
            msg += " using coordinates %s" % self.getObjectTag('inputCoordinates')
            msg += self.methodsVar.get('')
            methodsMsgs.append(msg)
        else:
            methodsMsgs.append("Set of Subtomograms not ready yet")
        if self.downFactor.get() != 1:
            methodsMsgs.append("Subtomograms downsample by factor %d."
                               % self.downFactor.get())
        return methodsMsgs

    def _summary(self):
        summary = []
        summary.append("Tomogram source: %s"
                       % self.getEnumText("tomoSource"))
        if self.getOutputsSize() >= 1:
            summary.append("Particle box size: %s" % self.boxSize.get())
            summary.append("Subtomogram extracted: %s" %
                           self.inputCoordinates.get().getSize())
        else:
            summary.append("Output subtomograms not ready yet.")
        return summary


# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pwem.wizards import EmWizard

from dynamo.protocols import DynamoExtraction


# =============================================================================
# EXTRACTION
# =============================================================================


class DynamoExtractionWizard(EmWizard):
    _targets = [(DynamoExtraction, ['boxSize'])]

    def show(self, form):
        DynamoExtractProt = form.protocol
        inputCoordinates = DynamoExtractProt.inputCoordinates.get()
        if not inputCoordinates:
            print('You must specify input coordinates')
            return

        aux = inputCoordinates.getBoxSize()
        if not aux % 2 == 0:
            aux += 1
        boxSize = aux
        if not boxSize:
            print('These coordinates do not have box size. Please, enter box size manually.')
            return

        if DynamoExtractProt.downFactor.get() != 1:
            boxSize = float(boxSize/DynamoExtractProt.downFactor.get())

        form.setVar('boxSize', boxSize)


# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Estrella Fernandez Gimenez (me.fernandez@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from .test_dynamo import *


# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Estrella Fernandez Gimenez (me.fernandez@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os import rename, remove
from os.path import join
from shutil import copy
from pwem import Domain
from pwem.objects.data import Volume, VolumeMask
from pyworkflow.object import Set
from pyworkflow.protocol.params import PointerParam, BooleanParam, IntParam, StringParam, FloatParam, LEVEL_ADVANCED
from pyworkflow.utils.path import makePath
from dynamo import Plugin
from dynamo.convert import writeVolume, writeSetOfVolumes, writeDynTable, readDynTable
ProtTomoSubtomogramAveraging = Domain.importFromPlugin("tomo.protocols.protocol_base", "ProtTomoSubtomogramAveraging")
AverageSubTomogram = Domain.importFromPlugin("tomo.objects", "AverageSubTomogram")
SetOfSubTomograms = Domain.importFromPlugin("tomo.objects", "SetOfSubTomograms")
SubTomogram = Domain.importFromPlugin("tomo.objects", "SubTomogram")


class DynamoSubTomoMRA(ProtTomoSubtomogramAveraging):
    """ This protocol will align subtomograms using Dynamo MRA Subtomogram Averaging"""

    _label = 'Subtomogram alignment'

    def __init__(self, **args):
        ProtTomoSubtomogramAveraging.__init__(self, **args)

    # --------------------------- DEFINE param functions ------------------------

    def _defineParams(self, form):
        form.addSection(label='Input subtomograms')
        form.addParam('inputVolumes', PointerParam, pointerClass="SetOfSubTomograms", label='Set of subtomograms',
                      help="Set of subtomograms to align with dynamo")
        form.addParam('sym', StringParam, default='c1', label='Symmetry group',
                      help="Specify the article's symmetry. Symmetrization is applied at the beginning of the round to "
                           "the input reference.")
        form.addParam('numberOfIters', IntParam, label='Iterations', default=5, help="Number of iterations per round")
        form.addParam('pca', BooleanParam, label='Perform PCA', default=False,
                      help="If selected, principal component analysis alignment is performed")

        form.addSection(label='Templates')
        form.addParam('mra', BooleanParam, label='Perform MRA', default=False,
                      help="If selected, multi-reference alignment (MRA) is performed")
        form.addParam('generateTemplate', BooleanParam, default=False, label='Generate reference template(s):',
                      help="Generate a reference template based on parameters")
        form.addParam('nref', IntParam, label='Number of references for MRA', default=2,
                      condition="mra and generateTemplate",
                      help="Number of references for multi-reference alignment (MRA)")
        form.addParam('templateRef', PointerParam, label="Template", condition="not generateTemplate",
                      pointerClass='Volume, SetOfVolumes',
                      help='The size of the template should be equal or smaller than the size of the particles. If you '
                           'pass a single file in multireference modus (MRA), Dynamo will just made copies of it.')
        form.addParam('useRandomTable', BooleanParam, label='Use a random table',
                      default=False, condition="generateTemplate", help="Rotate particles randomly")
        form.addParam('compensateMissingWedge', BooleanParam, label='Compensate for missing wedge', default=False,
                      condition="generateTemplate", help="")

        form.addSection(label='Masks')
        form.addParam('mask', PointerParam, label="Alignment mask", pointerClass='VolumeMask', allowsNull=True,
                      help='Needs to have the same dimmensionality as the template. It does NOT need to be smoothened. '
                           'A hard mask (only 0 and 1 values) is expected for the Roseman"s-like normalization')
        form.addParam('cmask', PointerParam, label="Classification mask", pointerClass='VolumeMask', allowsNull=True,
                      help='Needs to have the same dimmensionality as the template. It does NOT need to be smoothened. '
                           'A hard mask (only 0 and 1 values) is expected for the Roseman"s-like normalization')
        form.addParam('fmask', PointerParam, label="Fourier mask on template", pointerClass='Volume, SetOfVolumes',
                      allowsNull=True,
                      help='The fmask indicates which fourier coefficients are present at the starting reference volume'
                           '. The file should contain only ones or zeros.')
        form.addParam('smask', PointerParam, label="FSC mask", pointerClass='VolumeMask', allowsNull=True,
                      help='A direct space mask that will be imposed onto any couple of volumes when computing their '
                           'FSC (smask)')

        form.addSection(label='Angular scanning')
        form.addParam('cr', IntParam, label='Cone range', default=360,
                      help="The first two Euler angles are used to define the orientation of the vertical axis of the "
                           "protein. First Euler angle (tdrot) rotates the template around its z axis. Second Euler "
                           "angle (tilt) rotates the template around its x axis.Dynamo scans for this axis inside a "
                           "cone: The 'cone_range' parameter defines the angular aperture of this cone.360 degrees is "
                           "thus the value for a global scan. To skip the part of the angular search that looks for "
                           "orientations, you have to set 'cone range' to 0, and 'cone_sampling' to 1")
        form.addParam('cs', IntParam, label='Cone sampling', default=60,
                      help="This parameter expresses the discretization inside this cone. The sampling is given in "
                           "degrees, and corresponds to a representative angular distance between two neighboring "
                           "orientations inside the cone.")
        form.addParam('cf', IntParam, label='Cone flip', default=0, expertLevel=LEVEL_ADVANCED,
                      help="Generates a mirrored scanning geometry: the 'cone' of directions is complemented with the "
                           "diametrally oposed cone. This is useful when averaging elongated particles in the case in "
                           "which the direction of each one is not certain, i.e., the initial table catches the overall"
                           " orientation of each particle, but it is not certain on which end is the 'head' and which "
                           "is the 'tail', so that the refinement should allow 'flippling' the particles (but still "
                           "produce a scanning set of angles concentrated along the axis of the particle). 0:  No "
                           "inversion of the cone (default!) 1:  The cone is inverted only for the coarsest level of "
                           "the multigrid refinement 2:  The cone is inverted for all refinement levels")
        form.addParam('ccp', IntParam, label='Cone check peak', default=0, expertLevel=LEVEL_ADVANCED,
                      help="Controls peak quality inside the scanned directions. Useful when the particles are in close"
                           " contact with some feature of high intensity. Ensures that only angles yielding a real peak"
                           " will be considered. If the angle that yields the maximum of CC is in the boundary, the "
                           "particle will get aligned with its original alignment parameters. Values: 0: normal "
                           "behaviour (no peak quality control, default) any integer  : degrees that define the peak. A"
                           " maximum occurring within this distance to the boundary of the defined cone will be "
                           "discarded.")
        form.addParam('inplane_range', IntParam, label='Inplane rotation range', default=360,
                      help='The third Euler angle ("narot") defines rotations about the new axis of the reference. 360 '
                           'degrees is the value for a global scan. To skip the part of the angular search that looks '
                           'for azimuthal rotations , you have to set 1)  "inplane range" to zero, and 2)  '
                           '"inplane_sampling" to 1. Likewise,  to skip the part of the angular search that looks for '
                           'orientations, you have to manipulate the "cone" parameters: 1)  "cone range" to zero, and 2'
                           ') "cone_sampling" to 1.')
        form.addParam('inplane_sampling', IntParam, label='Inplane rotation sampling', default=45,
                      help='Parameter "inplane_sampling" is associated with the "narot" angle (New Axis ROTation). 1) '
                           'the axis of the template is rotated to a new orientation (defined by "tdrot" and "tilt"). 2'
                           ') the rotated template rotates again on its new axis (an "inplane" rotation). "inplane_'
                           'sampling" defines the angular interval (in degrees) between two of these inplane rotations')
        form.addParam('inplane_flip', IntParam, label='Inplane flip', default=0, expertLevel=LEVEL_ADVANCED,
                      help='Flips the set of inplane rotations. The set of inplane rotations to scan will be the '
                           'original set plus the flipped orientations.This is useful when the particles have a '
                           'directionality, but it is not very well defined. Values 0  :  no flip (default) 1  :  flips'
                           ' the coarsest level  in the multilevel grid   2  :  flips the full set (all levels).')
        form.addParam('inplane_check_peak', IntParam, label='Inplane check peak', default=0,
                      expertLevel=LEVEL_ADVANCED, help="Controls peak quality along the inplane rotation. Useful when "
                                                       "the particles are in close contact with some feature of high "
                                                       "intensity. Ensures that only angles yielding a real peak will "
                                                       "be considered. If the angle that yields the maximum of CC is in"
                                                       " the boundary, the particle will get aligned with its original "
                                                       "alignment parameters. Values: 0  : normal behaviour (no peak "
                                                       "quality control, default) any integer  : degrees that define "
                                                       "the peak. A maximum occurring within this distance to the "
                                                       "boundary of the defined range for inplane rotations will be "
                                                       "discarded.")
        form.addParam('rf', IntParam, label='Refine', default=5, expertLevel=LEVEL_ADVANCED,
                      help="How many refinement iterations are carried out on each single particle. This refinement "
                           "when comparing rotations of the reference against the data, takes the best orientation and "
                           "looks again with a finer sampling. The sampling in the refined search will be half of the "
                           "sampling used in the original one.  The range of the refined search encompasses all the "
                           "orientations that neighobur the best orientation found in the original search.")
        form.addParam('rff', IntParam, label='Refine factor', default=2, expertLevel=LEVEL_ADVANCED,
                      help="Controls the size of the angular neighborhood during the local refinement of the angular "
                           "grid.")

        form.addSection(label='Threshold')
        form.addParam('separation', IntParam, label='Separation in tomogram', default=0,
                      help='When tuned to  positive number, it will check the relative positions (positions in the '
                           'tomogram+shifts) of all the particles in each tomogram separately. Whenever two particles '
                           'are closer together than "separation_in_tomogram", only the particle with the higher '
                           'correlation will stay.')
        form.addParam('threshold', FloatParam, label='Threshold', default=0.2,
                      help='Different thresholding policies can be used in order to select which particles are averaged'
                           ' in view of their CC (cross correlation value) . The value of the thresholding parameter '
                           'defined here  will be interpreted differently depending on the "threshold_modus"')
        form.addParam('thresholdMode', IntParam, label='Threshold modus', default=1,
                      help='Possible values of the thresholding policy "threshold_modus": 0: no thesholding policy 1: '
                           'THRESHOLD is an absolute threshold (only particles with CC above this value are selected).'
                           ' 2: efective threshold = mean(CC) * THRESHOLD. 3: efective threshold = mean(CC) +std(CC)*'
                           'THRESHOLD. 4: THRESHOLD is the total number of particle (ordered by CC ). 5: THRESHOLD '
                           'ranges between 0 and 1  and sets the fraction of particles. * 11,21,31,34,41,51: select the'
                           ' same particles as 1,2,3,4 or 5, BUT non selected particles will be excluded: - from '
                           'averaging in the present iteration,and - ALSO from alignment in the next iteration (unlike '
                           '1,2,3,4,5).')
        form.addParam('threshold2', FloatParam, label='Second threshold', default=0.2, expertLevel=LEVEL_ADVANCED,
                      help="Thresholding II is operated against the average produced by the particles that survived the"
                           " first thresholding.")
        form.addParam('thresholdMode2', IntParam, label='Second threshold modus', default=0, expertLevel=LEVEL_ADVANCED,
                      help=" ")
        form.addParam('ccmatrix', BooleanParam, label='Compute  cross-correlation matrix', default=False,
                      help="Computation of a Cross-Correlation matrix among the aligned particles.")
        form.addParam('ccmatrixType', StringParam, label='Cross-correlation matrix type', default='align',
                      condition="ccmatrix", expertLevel=LEVEL_ADVANCED,
                      help="string with three characters, each position controling a different aspect: thresholding, "
                           "symmetrization, compensation")
        form.addParam('ccmatrixBatch', IntParam, label='Cross-correlation matrix batch', default=128,
                      expertLevel=LEVEL_ADVANCED,
                      help="Number of particles to be kept in memory simultaneously during the computation of the "
                           "ccmatrix. The larger this number, the more efficient the algorithm performance, as more "
                           "computations can be kept for reuse.However, trying to keep all the particles in memory can "
                           "lead to saturate it,blocking the CPU. Additionally, a small batch allows to divide the "
                           "matrix in more blocks. This might be useful in parallel computations.")
        form.addParam('low', IntParam, label='Low frequency', default=32,
                      expertLevel=LEVEL_ADVANCED, help='Cut off frequency for low pass filtering')
        form.addParam('high', IntParam, label='High frequency', default=2,
                      expertLevel=LEVEL_ADVANCED, help='Cut off frequency for high pass filtering')
        form.addParam('lim', StringParam, label='Area search', default='4 4 4', expertLevel=LEVEL_ADVANCED,
                      help='Restricts the search area to an ellipsoid centered and oriented in the last found position.'
                           ' The three parameters are the semiaxes of the ellipsoid. If a single parameter is '
                           'introduced, the ellipsoid collapses into a sphere. If no restriction should be imposed, put'
                           ' a zero on the "area search modus" parameter')
        form.addParam('limm', IntParam, label='Area search modus', default=0, expertLevel=LEVEL_ADVANCED,
                      help='States how exactly the shifts (area search) will be interpreted 0:  no limitations (can '
                           'easily produce artifacts if the initial reference is bad) 1:  limits are understood from '
                           'the center of the particle cube. 2:  limits are understood from the previous estimation on '
                           'the particle position (i.e., the shifts available in the table) With this option, the '
                           'origin of the shifts changes at every iteration. 3:  limits are understood from the '
                           'estimation provided for the first iteration of the round. The origin of the shifts will '
                           'change at each round. 4:  limits are understood from the estimation provided for the first '
                           'iteration')

    # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        self._insertFunctionStep('convertInputStep')
        self._insertFunctionStep('alignStep')
        self._insertFunctionStep('createOutputStep')
        self._insertFunctionStep('closeSetsStep')

    # --------------------------- STEPS functions -------------------------------

    def convertInputStep(self):
        fnDirData = self._getExtraPath("data")
        makePath(fnDirData)
        fnRoot = join(fnDirData, "particle_")
        inputVols = self.inputVolumes.get()
        writeSetOfVolumes(inputVols, fnRoot, 'id')
        pcaInt = int(self.pca.get())
        mraInt = int(self.mra.get())
        ccmatrixInt = int(self.ccmatrix.get())
        dim, _, _ = self.inputVolumes.get().getDimensions()
        self.projName = 'dynamoAlignmentProject'
        fnTable = self._getExtraPath("initial.tbl")
        fhTable = open(fnTable, 'w')
        writeDynTable(fhTable, inputVols)
        fhTable.close()
        fhCommands = open(self._getExtraPath("commands1.doc"), 'w')
        content = "dcp.new('%s','data','data','gui',0);" % self.projName + \
                  "dvput('%s', 'dim', '%s');" % (self.projName, dim) + \
                  "dvput('%s', 'sym', '%s');" % (self.projName, self.sym) + \
                  "dvput('%s', 'ite', '%s');" % (self.projName, self.numberOfIters) + \
                  "dvput('%s', 'mra', %s);" % (self.projName, mraInt) + \
                  "dvput('%s', 'pcas', %d);" % (self.projName, pcaInt) + \
                  "dvput('%s', 'cr', '%s');" % (self.projName, self.cr) + \
                  "dvput('%s', 'cs', '%s');" % (self.projName, self.cs) + \
                  "dvput('%s', 'cf', '%s');" % (self.projName, self.cf) + \
                  "dvput('%s', 'ccp', '%s');" % (self.projName, self.ccp) + \
                  "dvput('%s', 'rf', '%s');" % (self.projName, self.rf) + \
                  "dvput('%s', 'rff', '%s');" % (self.projName, self.rff) + \
                  "dvput('%s', 'ir', '%s');" % (self.projName, self.inplane_range) + \
                  "dvput('%s', 'is', '%s');" % (self.projName, self.inplane_sampling) + \
                  "dvput('%s', 'if', '%s');" % (self.projName, self.inplane_flip) + \
                  "dvput('%s', 'icp', '%s');" % (self.projName, self.inplane_check_peak) + \
                  "dvput('%s', 'thr', %s);" % (self.projName, self.threshold) + \
                  "dvput('%s', 'thrm', %s);" % (self.projName, self.thresholdMode) + \
                  "dvput('%s', 'thr2', %s);" % (self.projName, self.threshold2) + \
                  "dvput('%s', 'thr2m', %s);" % (self.projName, self.thresholdMode2) + \
                  "dvput('%s', 'ccms', %s);" % (self.projName, ccmatrixInt) + \
                  "dvput('%s', 'ccmt', '%s');" % (self.projName, self.ccmatrixType) + \
                  "dvput('%s', 'batch', '%s');" % (self.projName, self.ccmatrixBatch) +  \
                  "dvput('%s', 'stm', '%s');" % (self.projName, self.separation) + \
                  "dvput('%s', 'low', '%s');" % (self.projName, self.low) + \
                  "dvput('%s', 'high', '%s');" % (self.projName, self.high) + \
                  "dvput('%s', 'lim', '%s');" % (self.projName, self.lim) + \
                  "dvput('%s', 'limm', '%s');" % (self.projName, self.limm)

        template = self.templateRef.get()
        fmask = self.fmask.get()
        if template is not None:
            if isinstance(template, Volume):
                writeVolume(template, join(self._getExtraPath(), 'template'))
                content += "dvput('%s', 'template', 'template.mrc');" % self.projName
                content += "dvput('%s', 'table', 'initial.tbl');" % self.projName
            else:
                makePath(self._getExtraPath('templates'))
                writeSetOfVolumes(template, join(self._getExtraPath(), 'templates/template_initial_ref_'), 'ix')
                for ix, template1 in enumerate(self.templateRef.get().iterItems()):
                    copy(join(self._getExtraPath(), 'initial.tbl'),
                         join(self._getExtraPath(), 'templates/table_initial_ref_%03d.tbl' % int(ix+1)))
                content += "dvput('%s', 'template', 'templates');" % self.projName
                content += "dvput('%s', 'table', 'templates');" % self.projName
                content += "dvput('%s', 'nref', '%d');" % (self.projName, self.nref)
                if fmask is None:
                        makePath(self._getExtraPath('fmasks'))
                        dimsf = inputVols.getFirstItem().getDim()
                        sizef = template.getSize()
                        content += "dynamo_write_multireference(ones(%d,%d,%d),'fmask','fmasks','refs',1:%d);" \
                                   % (dimsf[0], dimsf[1], dimsf[2], sizef)
                        content += "dvput('%s', 'fmask', 'fmasks');" % self.projName
        else:
            if not self.mra.get():
                if self.useRandomTable.get():
                    content += "dynamo_table_perturbation('initial.tbl','pshift',0,'paxis',0,'pnarot',360,'o'," \
                               "'initial.tbl');"
                if self.compensateMissingWedge.get():
                    content += "dynamo_table_randomize_azimuth('initial.tbl','o','initial.tbl');"
                content += "dynamo_average('data','table','initial.tbl','o','template.mrc');"
                content += "dvput('%s', 'template', 'template.mrc');" % self.projName
                content += "dvput('%s', 'table', 'initial.tbl');" % self.projName

            else:
                makePath(self._getExtraPath('templates'))
                for ix, template in enumerate(range(self.nref.get())):
                    if self.useRandomTable.get():
                        content += "dynamo_table_perturbation('initial.tbl','pshift',0,'paxis',0,'pnarot',360,'o'," \
                                   "'initial.tbl');"
                    if self.compensateMissingWedge.get():
                        content += "dynamo_table_randomize_azimuth('initial.tbl','o','initial.tbl');"
                    content += "dynamo_average('data','table','initial.tbl','o'," \
                               "'templates/template_initial_ref_%03d.mrc');" % int(ix+1)
                    copy(join(self._getExtraPath(), 'initial.tbl'),
                         join(self._getExtraPath(), 'templates/table_initial_ref_%03d.tbl' % int(ix+1)))
                content += "dvput('%s', 'template', 'templates');" % self.projName
                content += "dvput('%s', 'table', 'templates');" % self.projName
                content += "dvput('%s', 'nref', '%d');" % (self.projName, self.nref)
                if fmask is None:
                        makePath(self._getExtraPath('fmasks'))
                        dimsf = inputVols.getFirstItem().getDim()
                        sizef = self.nref.get()
                        content += "dynamo_write_multireference(ones(%d,%d,%d),'fmask','fmasks','refs',1:%d);" \
                                   % (dimsf[0], dimsf[1], dimsf[2], sizef)
                        content += "dvput('%s', 'fmask', 'fmasks');" % self.projName

        if self.mask.get() is not None:
            writeVolume(self.mask.get(), join(self._getExtraPath(), 'mask'))
            content += "dvput('%s', 'mask', 'mask.mrc');" % self.projName
        if self.cmask.get() is not None:
            writeVolume(self.cmask.get(), join(self._getExtraPath(), 'cmask'))
            content += "dvput('%s', 'cmask', 'cmask.mrc');" % self.projName
        if fmask is not None:
            if isinstance(fmask, Volume) or isinstance(fmask, SubTomogram) or isinstance(fmask, VolumeMask):
                writeVolume(fmask, join(self._getExtraPath(), 'fmask'))
                content += "dvput('%s', 'fmask', 'fmask.mrc');" % self.projName
            else:
                makePath(self._getExtraPath('fmasks'))
                writeSetOfVolumes(fmask, join(self._getExtraPath(), 'fmasks/fmask_initial_ref_'), 'ix')
                content += "dvput('%s', 'fmask', 'fmasks');" % self.projName
        if self.smask.get() is not None:
            writeVolume(self.smask.get(), join(self._getExtraPath(), 'smask'))
            content += "dvput('%s', 'smask', 'smask.mrc');" % self.projName

        content += "dynamo_data_format('templates/template_initial_ref_*.mrc'," \
                   "'templates','modus','convert','extension','.em');"

        fhCommands.write(content)
        fhCommands.close()

        Plugin.runDynamo(self, 'commands1.doc', cwd=self._getExtraPath())

        if mraInt == 1:
            if self.generateTemplate.get() == 0:
                numberOfrefs = self.templateRef.get().getSize()
            else:
                numberOfrefs = self.nref.get()
            for iref in range(numberOfrefs):
                rename(join(self._getExtraPath(), 'templates/particle_%05d.em' % int(iref+1)),
                       join(self._getExtraPath(), 'templates/template_initial_ref_%03d.em' % int(iref+1)))
                remove(join(self._getExtraPath(), 'templates/template_initial_ref_%03d.mrc' % int(iref+1)))

    def alignStep(self):
        fhCommands2 = open(self._getExtraPath("commands2.doc"), 'w')
        content2 = "dvcheck('%s');" % self.projName
        content2 += "dvunfold('%s');dynamo_execute_project %s" % (self.projName, self.projName)
        fhCommands2.write(content2)
        fhCommands2.close()
        Plugin.runDynamo(self, 'commands2.doc', cwd=self._getExtraPath())

    def createOutputStep(self):
        self.projName = 'dynamoAlignmentProject'
        niters = int(self.numberOfIters.get())
        if self.mra.get() == 0:
            self.subtomoSet = self._createSetOfSubTomograms()
            inputSet = self.inputVolumes.get()
            self.subtomoSet.copyInfo(inputSet)
            self.fhTable = open(self._getExtraPath('%s/results/ite_%04d/averages/refined_table_ref_001_ite_%04d.tbl')
                                % (self.projName, niters, niters), 'r')
            self.subtomoSet.copyItems(inputSet, updateItemCallback=self._updateItem)
            self.fhTable.close()
            averageSubTomogram = AverageSubTomogram()
            averageSubTomogram.setFileName(
                self._getExtraPath('%s/results/ite_%04d/averages/average_ref_001_ite_%04d.em')
                % (self.projName, niters, niters))
            averageSubTomogram.setSamplingRate(inputSet.getSamplingRate())
            self._defineOutputs(outputSubtomograms=self.subtomoSet)
            self._defineSourceRelation(self.inputVolumes, self.subtomoSet)
            self._defineOutputs(averageSubTomogram=averageSubTomogram)
            self._defineSourceRelation(self.inputVolumes, averageSubTomogram)
        else:
            fhSurvivRefs = open(self._getExtraPath('%s/results/ite_%04d/currently_surviving_references_ite_%04d.txt')
                                % (self.projName, niters, niters), 'r')
            nline = next(fhSurvivRefs).rstrip()
            for nref in range(len(nline.split())):
                ref = int(nline.split()[nref])
                subtomoSet = self._createSetOfSubTomograms()
                inputSet = self.inputVolumes.get()
                subtomoSet.copyInfo(inputSet)
                self.fhTable = open(
                    self._getExtraPath('%s/results/ite_%04d/averages/refined_table_ref_%03d_ite_%04d.tbl') %
                    (self.projName, niters, ref, niters), 'r')
                subtomoSet.copyItems(inputSet, updateItemCallback=self._updateItem)
                self.fhTable.close()
                averageSubTomogram = AverageSubTomogram()
                averageSubTomogram.setFileName(
                    self._getExtraPath('%s/results/ite_%04d/averages/average_ref_%03d_ite_%04d.em')
                    % (self.projName, niters, ref, niters))
                averageSubTomogram.setSamplingRate(inputSet.getSamplingRate())

                name = 'outputSubtomogramsRef%s' % str(ref)
                args = {}
                args[name] = subtomoSet
                subtomoSet.setStreamState(Set.STREAM_OPEN)
                self._defineOutputs(**args)
                self._defineSourceRelation(inputSet, subtomoSet)

                name2 = 'AverageRef%s' % str(ref)
                args2 = {}
                args2[name2] = averageSubTomogram
                self._defineOutputs(**args2)
                self._defineSourceRelation(inputSet, averageSubTomogram)

    def closeSetsStep(self):
        if self.mra.get() == True:
            for outputset in self._iterOutputsNew():
                if isinstance(outputset, SetOfSubTomograms):
                    outputset[1].setStreamState(Set.STREAM_CLOSED)
            self._store()

    # --------------------------- UTILS functions --------------------------------
    def _updateItem(self, item, row):
        readDynTable(self, item)

    # --------------------------- INFO functions --------------------------------
    def _validate(self):
        validateMsgs = []
        if self.mra.get() == True:
            if self.generateTemplate.get():
                if self.nref.get() <= 1:
                    validateMsgs.append('If MRA is selected, number of references should be greater than one.')
            elif isinstance(self.templateRef.get(), Volume):
                validateMsgs.append('If MRA is selected, number of references should be greater than one.')
            elif self.templateRef.get().getSize() <= 1:
                validateMsgs.append('If MRA is selected, number of references should be greater than one.')
        return validateMsgs

    def _summary(self):
        summary = []
        summary.append("Input subtomograms: %d" % self.inputVolumes.get().getSize())
        if self.fmask.get() is not None:
            if isinstance(self.fmask.get(), Volume) or isinstance(self.fmask.get(), SubTomogram):
                summary.append("Input fmask: %s" % self.fmask.get())
            else:
                summary.append("Input fmasks: %d" % self.fmask.get().getSize())
        else:
            summary.append("Fmask(s) generated")
        if self.mra.get() == True:
            summary.append("Perform MRA with %s references" % self.nref.get())
        else:
            summary.append("No mra")
        if self.generateTemplate.get():
            summary.append("Template(s) generated")
        else:
            if isinstance(self.templateRef.get(), Volume) or isinstance(self.templateRef.get(), SubTomogram):
                summary.append("Provided template: %s" % self.templateRef.get())
            else:
                summary.append("Provided templates: %d" % self.templateRef.get().getSize())
        return summary

    def _methods(self):
        methods = []
        methods.append(
            'We aligned %d subtomograms from %s using Dynamo Subtomogram averaging.'
            % (self.inputVolumes.get().getSize(), self.getObjectTag('inputVolumes')))
        return methods

    def _citations(self):
        return ['CASTANODIEZ2012139']
